#if defined(__e2k__) 
#ifndef DONT_USE_EML
/* inffast.c -- fast decoding
 * Copyright (c) 2015-2024 AO MCST
 * Copyright (C) 1995-2017 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

#include "zutil.h"
#include "inftrees.h"
#include "inflate.h"
#include "inffast.h"


#define POSTINC

#ifndef DONT_USE_EML

#include <eml.h>
#include <eml_owndefs.h>
#include <e2k_intrin.h>

#ifdef __ptr128__
#define EML_INIT(size)                \
    unsigned c1 = src[0], c2, c3, c4; \
    if (size > 1) c2 = src[1];        \
    if (size > 2) c3 = src[2];        \
    if (size > 3) c4 = src[3]
#else /* __ptr128__ */
#define EML_INIT(size)                \
    unsigned c1 = src[0], c2 = src[1], c3 = src[2], c4 = src[3]
#endif /* __ptr128__ */

/* заполнение dst длиной (len > 16) байт блоками по size (1 - 4) байт каждый */
#define EML_SPEC_MEMSET(dst, size, len) {                               \
    unsigned char * __restrict__ src = dst - size;                      \
    int i;                                                              \
    EML_INIT (size);                                                    \
                                                                        \
    if (size == 1) {                                                    \
_Pragma ("loop count (7)")                                              \
        for (i = 0; i < len; i)                                       \
            dst[i] = c1;                                                \
    }                                                                   \
    else if (size == 2) {                                               \
_Pragma ("loop count (3)")                                              \
        for (i = 0; i < (len - 1); i = 2) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        if (i < len) dst[i] = c1;                                       \
    }                                                                   \
    else if (size == 3) {                                               \
_Pragma ("loop count (2)")                                              \
        for (i = 0; i < (len - 2); i = 3) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
        }                                                               \
        if (i < (len - 1)) {                                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
        }                                                               \
        if (i < (len - 1)) {                                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        else if (i < len) dst[i] = c1;                                  \
    }                                                                   \
    else /* (size == 4) */ {                                            \
_Pragma ("loop count (2)")                                              \
        for (i = 0; i < (len - 3); i = 4) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
            dst[i  3] = c4;                                            \
        }                                                               \
        if (i < (len - 2)) {                                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
        }                                                               \
        else if (i < (len - 1)) {                                       \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        else if (i < len) dst[i] = c1;                                  \
    }                                                                   \
}

/* заполнение dst длиной (len > 16) байт блоками по size (1 - 4) байт каждый */
void ZLIB_INTERNAL eml_spec_memset (unsigned char *dst,
                                    int size,
                                    int len)
{
    eml_64u cc, * __restrict__ dp;
    unsigned char * __restrict__ src = dst - size;
    int i, head_len, body_len, tail_len;
#ifdef __ptr128__
    eml_32u c1 = src[0], c2, c3, c4;
    if (size > 1) c2 = src[1];
    if (size > 2) c3 = src[2];
    if (size > 3) c4 = src[3];
#else /* __ptr128__ */
    eml_32u c1 = src[0], c2 = src[1], c3 = src[2], c4 = src[3];
#endif /* __ptr128__ */

    /* для корректной работы основного кода len должна быть не меньше 8. */
    head_len = EML_BYTES_TO_ALIGN (dst, 8);
    body_len = (len - head_len) >> 3;
    tail_len = (len - head_len) & 7;

    if (size == 1) {
        cc = (c1 << 8) | c1;
        cc = e2k_pshufh (cc, 0); /* 0000 */
    }
    else if (size == 2) {
        cc = (c2 << 8) | c1;
        cc = e2k_pshufh (cc, 0); /* 0000 */
    }
    else if (size == 3) {
        eml_64u cc0, cc1, cc2;
        cc = (c3 << 16) | (c2 << 8) | c1;
        cc = (cc << 24) | cc;
        cc = (cc << 48) | cc;

#pragma loop count(3)
        for (i = 0; i < head_len; i) {
            dst[i] = cc;
            cc = e2k_pshufh (cc, 0x92);
        }

        dp = (eml_64u *) (dst  head_len);
        cc0 = cc;
        cc1 = e2k_pshufh (cc, 0x49);
        cc2 = e2k_pshufh (cc, 0x92);

#pragma unroll(2)               /* 1 ~T~K */
#pragma loop count(1000)
        for (i = 0; i < body_len; i) {
            dp[i] = cc0;
            cc = cc1;
            cc1 = cc2;
            cc2 = cc0;
            cc0 = cc;
        }

#pragma loop count(3)
        for (i = len - tail_len; i < len; i) {
            dst[i] = cc;
            cc = e2k_pshufh (cc, 0x92);
        }
        return;
    }
    else /* (size == 4) */ {
        cc = (c4 << 24) | (c3 << 16) | (c2 << 8) | c1;
        cc = e2k_pshufh (cc, 0x44); /* 1010 */
    }

#pragma loop count(3)
    for (i = 0; i < head_len; i) {
        dst[i] = cc;
        cc = e2k_scrd (cc, 8);
    }

    dp = (eml_64u *) (dst  head_len);

#pragma unroll(2)               /* 1 ~T~K */
#pragma loop count(1000)
    for (i = 0; i < body_len; i) {
        dp[i] = cc;
    }

#pragma loop count(3)
    for (i = len - tail_len; i < len; i) {
        dst[i] = cc;
        cc = e2k_scrd (cc, 8);
    }
}

#endif /* DONT_USE_EML */


#ifdef ASMINF
#  pragma message("Assembler code may have bugs -- use at your own risk")
#else

/*
   Decode literal, length, and distance codes and write out the resulting
   literal and match bytes until either not enough input or output is
   available, an end-of-block is encountered, or a data error is encountered.
   When large enough input and output buffers are supplied to inflate(), for
   example, a 16K input buffer and a 64K output buffer, more than 95% of the
   inflate execution time is spent in this routine.

   Entry assumptions:

        state->mode == LEN
        strm->avail_in >= 6
        strm->avail_out >= 258
        start >= strm->avail_out
        state->bits < 8

   On return, state->mode is one of:

        LEN -- ran out of enough output space or enough available input
        TYPE -- reached end of block code, inflate() to interpret next block
        BAD -- error in block data

   Notes:

    - The maximum input bits used by a length/distance pair is 15 bits for the
      length code, 5 bits for the length extra, 15 bits for the distance code,
      and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
      Therefore if strm->avail_in >= 6, then there is enough input to avoid
      checking for available input while decoding.

    - The maximum bytes that a single length/distance pair can output is 258
      bytes, which is the maximum length that can be coded.  inflate_fast()
      requires strm->avail_out >= 258 for each loop to avoid checking for
      output space.
 */
void ZLIB_INTERNAL inflate_fast(strm, start)
z_streamp strm;
unsigned start;         /* inflate()'s starting value for strm->avail_out */
{
    struct inflate_state FAR * __restrict__ state;
    z_const unsigned char FAR * __restrict__ in;      /* local strm->next_in */
    z_const unsigned char FAR * __restrict__ last;    /* have enough input while in < last */
    unsigned char FAR * __restrict__ out;     /* local strm->next_out */
    unsigned char FAR * __restrict__ beg;     /* inflate()'s initial strm->next_out */
    unsigned char FAR * __restrict__ end;     /* while out < end, enough space available */
#ifdef INFLATE_STRICT
    unsigned dmax;              /* maximum distance from zlib header */
#endif
    unsigned wsize;             /* window size or zero if not using window */
    unsigned whave;             /* valid bytes in the window */
    unsigned wnext;             /* window write index */
    unsigned char FAR * __restrict__ window;  /* allocated sliding window, if wsize != 0 */
    unsigned long hold;         /* local strm->hold */
    unsigned bits;              /* local strm->bits */
    code const FAR * __restrict__ lcode;      /* local strm->lencode */
    code const FAR * __restrict__ dcode;      /* local strm->distcode */
    unsigned lmask;             /* mask for first level of length codes */
    unsigned dmask;             /* mask for first level of distance codes */
    code here;                  /* retrieved table entry */
    unsigned op;                /* code bits, operation, extra bits, or */
                                /*  window position, window bytes to copy */
    unsigned len;               /* match length, unused bytes */
    unsigned dist;              /* match distance */
    unsigned char FAR * __restrict__ from;    /* where to copy match from */

    /* copy state to local variables */
    state = (struct inflate_state FAR *)strm->state;
    in = strm->next_in;
    last = in  (strm->avail_in - 5);
    out = strm->next_out;
    beg = out - (start - strm->avail_out);
    end = out  (strm->avail_out - 257);
#ifdef INFLATE_STRICT
    dmax = state->dmax;
#endif
    wsize = state->wsize;
    whave = state->whave;
    wnext = state->wnext;
    window = state->window;
    hold = state->hold;
    bits = state->bits;
    lcode = state->lencode;
    dcode = state->distcode;
    lmask = (1U << state->lenbits) - 1;
    dmask = (1U << state->distbits) - 1;

    /* decode literals and length/distances until end-of-block or not enough
       input data or output space */
//#pragma loop count (10000)
    do {
        if (bits < 15) {
            hold = (unsigned long)(*in) << bits;
            bits = 8;
            hold = (unsigned long)(*in) << bits;
            bits = 8;
        }
        here = lcode[hold & lmask];
      dolen:
        op = (unsigned)(here.bits);
        hold >>= op;
        bits -= op;
        op = (unsigned)(here.op);
        if (op == 0) {                          /* literal */
            Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
                    "inflate:         literal '%c'\n" :
                    "inflate:         literal 0x%02x\n", here.val));
            *out = (unsigned char)(here.val);
        }
        else if (op & 16) {                     /* length base */
            len = (unsigned)(here.val);
            op &= 15;                           /* number of extra bits */
            if (op) {
                if (__builtin_expect ((bits < op), 0)) {
                    hold = (unsigned long)(*in) << bits;
                    bits = 8;
                }
                len = (unsigned)hold & ((1U << op) - 1);
                hold >>= op;
                bits -= op;
            }
            Tracevv((stderr, "inflate:         length %u\n", len));
            if (bits < 15) {
                hold = (unsigned long)(*in) << bits;
                bits = 8;
                hold = (unsigned long)(*in) << bits;
                bits = 8;
            }
            here = dcode[hold & dmask];
          dodist:
            op = (unsigned)(here.bits);
            hold >>= op;
            bits -= op;
            op = (unsigned)(here.op);
            if (op & 16) {                      /* distance base */
                dist = (unsigned)(here.val);
                op &= 15;                       /* number of extra bits */
                if (bits < op) {
                    hold = (unsigned long)(*in) << bits;
                    bits = 8;
                    if (__builtin_expect ((bits < op), 0)) {
                        hold = (unsigned long)(*in) << bits;
                        bits = 8;
                    }
                }
                dist = (unsigned)hold & ((1U << op) - 1);
#ifdef INFLATE_STRICT
                if (dist > dmax) {
                    strm->msg = (char *)"invalid distance too far back";
                    state->mode = BAD;
                    break;
                }
#endif
                hold >>= op;
                bits -= op;
                Tracevv((stderr, "inflate:         distance %u\n", dist));
                op = (unsigned)(out - beg);     /* max distance in output */
                if (dist > op) {                /* see if copy from window */
                    op = dist - op;             /* distance back in window */
                    if (__builtin_expect ((op > whave), 0)) {
                        if (state->sane) {
                            strm->msg =
                                (char *)"invalid distance too far back";
                            state->mode = BAD;
                            break;
                        }
#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
                        if (len <= op - whave) {
                            do {
                                *out = 0;
                            } while (--len);
                            continue;
                        }
                        len -= op - whave;
                        do {
                            *out = 0;
                        } while (--op > whave);
                        if (op == 0) {
                            from = out - dist;
                            do {
                                *out = *from;
                            } while (--len);
                            continue;
                        }
#endif
                    }
                    from = window;
                    if (__builtin_expect ((wnext == 0), 0)) {           /* very common case */
                        from = wsize - op;
                        if (op < len) {         /* some from window */
                            len -= op;
                            do {
                                *out = *from;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
                    else if (wnext < op) {      /* wrap around window */
                        from = wsize  wnext - op;
                        op -= wnext;
                        if (__builtin_expect ((op < len), 0)) {         /* some from end of window */
                            len -= op;
#pragma loop count (3)
                            do {
                                *out = *from;
                            } while (--op);
                            from = window;
                            if (__builtin_expect ((wnext < len), 0)) {  /* some from start of window */
                                op = wnext;
                                len -= op;
                                do {
                                    *out = *from;
                                } while (--op);
                                from = out - dist;      /* rest from output */
                            }
                        }
                    }
                    else {                      /* contiguous in window */
                        from = wnext - op;
                        if (__builtin_expect ((op < len), 0)) {         /* some from window */
                            len -= op;
#pragma loop count (3)
                            do {
                                *out = *from;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
#pragma loop count (2)
                    while (len > 2) {
                        *out = *from;
                        *out = *from;
                        *out = *from;
                        len -= 3;
                    }
                    if (len) {
                        *out = *from;
                        if (len > 1)
                            *out = *from;
                    }
                }
                else {
#ifndef DONT_USE_EML
                    if (dist <= 4) { /* заполняем пачками по dist байт сразу и выставляем out */
                        if (len <= 16) {
                            EML_SPEC_MEMSET (out, ((int) dist), ((int) len));
                        }
                        else
                            eml_spec_memset (out, ((int) dist), (int) len);
                        out = len;
                    }
                    else {
#endif /* DONT_USE_EML */
                    from = out - dist;          /* copy direct from output */
//FIXME: if (len > dist) -> out и from пересекаются. Пока так работает.
#pragma unroll (1)
#pragma loop count (2) /* чтобы все было примитивно и никаких оптимизаций */
                        do {                        /* minimum length is three */
                            *out = *from;
                            *out = *from;
                            *out = *from;
                            len -= 3;
                        } while (len > 2);
                        if (len) {
                            *out = *from;
                            if (len > 1)
                                *out = *from;
                        }
#ifndef DONT_USE_EML
                    }
#endif /* DONT_USE_EML */
                }
            }
            else if (__builtin_expect (((op & 64) == 0), 1)) {          /* 2nd level distance code */
                here = dcode[here.val  (hold & ((1U << op) - 1))];
                goto dodist;
            }
            else {
                strm->msg = (char *)"invalid distance code";
                state->mode = BAD;
                break;
            }
        }
        else if (__builtin_expect (((op & 64) == 0), 1)) {              /* 2nd level length code */
            here = lcode[here.val  (hold & ((1U << op) - 1))];
            goto dolen;
        }
        else if (__builtin_expect ((op & 32), 1)) {                     /* end-of-block */
            Tracevv((stderr, "inflate:         end of block\n"));
            state->mode = TYPE;
            break;
        }
        else {
            strm->msg = (char *)"invalid literal/length code";
            state->mode = BAD;
            break;
        }
    } while (in < last && out < end);

    /* return unused bytes (on entry, bits < 8, so in won't go too far back) */
    len = bits >> 3;
    in -= len;
    bits -= len << 3;
    hold &= (1U << bits) - 1;

    /* update state and return */
    strm->next_in = in;
    strm->next_out = out;
    strm->avail_in = (unsigned)(in < last ? 5  (last - in) : 5 - (in - last));
    strm->avail_out = (unsigned)(out < end ?
                                 257  (end - out) : 257 - (out - end));
    state->hold = hold;
    state->bits = bits;
    return;
}

/*
   inflate_fast() speedups that turned out slower (on a PowerPC G3 750CXe):
   - Using bit fields for code structure
   - Different op definition to avoid & for extra bits (do & for table bits)
   - Three separate decoding do-loops for direct, window, and wnext == 0
   - Special case for distance > 1 copies to do overlapped load and store copy
   - Explicit branch predictions (based on measured branch probabilities)
   - Deferring match copy and interspersed it with decoding subsequent codes
   - Swapping literal/length else
   - Swapping window/direct else
   - Larger unrolled copy loops (three is about right)
   - Moving len -= 3 statement into middle of loop
 */

#endif /* !ASMINF */

#endif
#endif


#if defined (__sparc_v9__) 
#ifndef iDONT_USE_EML
/* inffast.c -- fast decoding
 * Copyright (C) 1995-2017 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

#include "zutil.h"
#include "inftrees.h"
#include "inflate.h"
#include "inffast.h"


#define POSTINC

#ifndef DONT_USE_EML

#include <eml.h>
#include <vis.h>

/* заполнение dst длиной (len > 16) байт блоками по size (1 - 4) байт каждый *
#define EML_SPEC_MEMSET(dst, size, len) {                               \
    unsigned char * __restrict__ src = dst - size;                      \
    eml_32u c1 = src[0], c2 = src[1], c3 = src[2], c4 = src[3];         \
    int i;                                                              \
                                                                        \
    if (size == 1) {                                                    \
_Pragma ("loop count (7)")                                              \
        for (i = 0; i < len; i)                                       \
            dst[i] = c1;                                                \
    }                                                                   \
    else if (size == 2) {                                               \
_Pragma ("loop count (3)")                                              \
        for (i = 0; i < (len - 1); i = 2) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        if (i < len) dst[i] = c1;                                       \
    }                                                                   \
    else if (size == 3) {                                               \
_Pragma ("loop count (2)")                                              \
        for (i = 0; i < (len - 2); i = 3) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
        }                                                               \
        if (i < (len - 1)) {                                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        else if (i < len) dst[i] = c1;                                  \
    }                                                                   \
    else /* (size == 4) */ {                                            \
_Pragma ("loop count (2)")                                              \
        for (i = 0; i < (len - 3); i = 4) {                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
            dst[i  3] = c4;                                            \
        }                                                               \
        if (i < (len - 2)) {                                            \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
            dst[i  2] = c3;                                            \
        }                                                               \
        else if (i < (len - 1)) {                                       \
            dst[i  0] = c1;                                            \
            dst[i  1] = c2;                                            \
        }                                                               \
        else if (i < len) dst[i] = c1;                                  \
    }                                                                   \
}

/* заполнение dst длиной (len > 16) байт блоками по size (1 - 4) байт каждый */
void ZLIB_INTERNAL eml_spec_memset (unsigned char *dst,
                                    int size,
                                    int len)
{
    eml_64f cc, * __restrict__ dp;
    unsigned char * __restrict__ src = dst - size, * __restrict__ dstend;
    eml_32u c1 = src[0], c2 = src[1], c3 = src[2], c4 = src[3];
    int i, emask;

     /* для корректной работы основного кода len должна быть не меньше 8. */

    if (size == 1) {
        c1 = (c1 << 8) | c1;
        c1 = (c1 << 16) | c1;
    }
    else if (size == 2) {
        c1 = (c1 << 8) | c2;
        c1 = (c1 << 16) | c1;
    }
    else if (size == 3) {
        eml_64f cc1, cc2, cc00, cc11, cc22;
        c1 = (c1 << 24) | (c2 << 16) | (c3 << 8) | c1;
        c2 = (c1 << 8) | c2;
        c3 = (c2 << 8) | c3;

        cc = vis_to_double (c1, c2);
        cc1 = vis_to_double (c3, c1);
        cc2 = vis_to_double (c2, c3);

        dp = (eml_64f *) ((eml_addr) dst & ~7);
        dstend = dst  len - 1;
        vis_alignaddr ((void *) (-(eml_addr) dst), 0);
        emask = vis_edge8 (dst, dstend);
        cc22 = vis_faligndata (cc2, cc);
        cc00 = vis_faligndata (cc, cc1);
        cc11 = vis_faligndata (cc1, cc2);
        if ((eml_addr) dst & 7)
            vis_pst_8 (cc22, dp, emask);
        i = (eml_addr) dp - (eml_addr) dst;

#pragma unroll(2)
#pragma loop count(1000)
        for (; i < len - 24; i = 24) {
            dp[0] = cc00;
            dp[1] = cc11;
            dp[2] = cc22;
            dp = 3;
        }
        if (i < (len - 8)) {
            *dp = cc00;

            if (i < (len - 16)) {
                *dp = cc11;
                cc00 = cc22;
            }
            else
                cc00 = cc11;
        }
        emask = vis_edge8 (dp, dstend);
        vis_pst_8 (cc00, dp, emask);
        return;
    }
    else /* (size == 4) */ {
        c1 = (c1 << 24) | (c2 << 16) | (c3 << 8) | c4;
    }

    if (size != 3) { /* size == 1, 2 ~I~I4 */
        cc = vis_to_double_dup (c1);
        dp = (eml_64f *) ((eml_addr) dst & ~7);
        dstend = dst  len - 1;
        vis_alignaddr ((void *) (-(eml_addr) dst), 0);
        emask = vis_edge8 (dst, dstend);
        cc = vis_faligndata (cc, cc);
        vis_pst_8 (cc, dp, emask);
        i = (eml_addr) dp - (eml_addr) dst;

#pragma unroll(4)
#pragma loop count(1000)
        for (; i < len - 8; i = 8) {
            *dp = cc;
        }
        emask = vis_edge8 (dp, dstend);
        vis_pst_8 (cc, dp, emask);
        return;
    }
    /* size == 3 */
}

#endif /* DONT_USE_EML */

#ifdef ASMINF
#  pragma message("Assembler code may have bugs -- use at your own risk")
#else

/*
   Decode literal, length, and distance codes and write out the resulting
   literal and match bytes until either not enough input or output is
   available, an end-of-block is encountered, or a data error is encountered.
   When large enough input and output buffers are supplied to inflate(), for
   example, a 16K input buffer and a 64K output buffer, more than 95% of the
   inflate execution time is spent in this routine.

   Entry assumptions:

        state->mode == LEN
        strm->avail_in >= 6
        strm->avail_out >= 258
        start >= strm->avail_out
        state->bits < 8

   On return, state->mode is one of:

        LEN -- ran out of enough output space or enough available input
        TYPE -- reached end of block code, inflate() to interpret next block
        BAD -- error in block data

   Notes:

    - The maximum input bits used by a length/distance pair is 15 bits for the
      length code, 5 bits for the length extra, 15 bits for the distance code,
      and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
      Therefore if strm->avail_in >= 6, then there is enough input to avoid
      checking for available input while decoding.

    - The maximum bytes that a single length/distance pair can output is 258
      bytes, which is the maximum length that can be coded.  inflate_fast()
      requires strm->avail_out >= 258 for each loop to avoid checking for
      output space.
 */
void ZLIB_INTERNAL inflate_fast(strm, start)
z_streamp strm;
unsigned start;         /* inflate()'s starting value for strm->avail_out */
{
    struct inflate_state FAR * __restrict__ state;
    z_const unsigned char FAR * __restrict__ in;      /* local strm->next_in */
    z_const unsigned char FAR * __restrict__ last;    /* have enough input while in < last */
    unsigned char FAR * __restrict__ out;     /* local strm->next_out */
    unsigned char FAR * __restrict__ beg;     /* inflate()'s initial strm->next_out */
    unsigned char FAR * __restrict__ end;     /* while out < end, enough space available */
#ifdef INFLATE_STRICT
    unsigned dmax;              /* maximum distance from zlib header */
#endif
    unsigned wsize;             /* window size or zero if not using window */
    unsigned whave;             /* valid bytes in the window */
    unsigned wnext;             /* window write index */
    unsigned char FAR * __restrict__ window;  /* allocated sliding window, if wsize != 0 */
    unsigned long hold;         /* local strm->hold */
    unsigned bits;              /* local strm->bits */
    code const FAR * __restrict__ lcode;      /* local strm->lencode */
    code const FAR * __restrict__ dcode;      /* local strm->distcode */
    unsigned lmask;             /* mask for first level of length codes */
    unsigned dmask;             /* mask for first level of distance codes */
    code here;                  /* retrieved table entry */
    unsigned op;                /* code bits, operation, extra bits, or */
                                /*  window position, window bytes to copy */
    unsigned len;               /* match length, unused bytes */
    unsigned dist;              /* match distance */
    unsigned char FAR * __restrict__ from;    /* where to copy match from */

    /* copy state to local variables */
    state = (struct inflate_state FAR *)strm->state;
    in = strm->next_in;
    last = in  (strm->avail_in - 5);
    out = strm->next_out;
    beg = out - (start - strm->avail_out);
    end = out  (strm->avail_out - 257);
#ifdef INFLATE_STRICT
    dmax = state->dmax;
#endif
    wsize = state->wsize;
    whave = state->whave;
    wnext = state->wnext;
    window = state->window;
    hold = state->hold;
    bits = state->bits;
    lcode = state->lencode;
    dcode = state->distcode;
    lmask = (1U << state->lenbits) - 1;
    dmask = (1U << state->distbits) - 1;

    /* decode literals and length/distances until end-of-block or not enough
       input data or output space */
//#pragma loop count (10000)
    do {
        if (bits < 15) {
            hold = (unsigned long)(*in) << bits;
            bits = 8;
            hold = (unsigned long)(*in) << bits;
            bits = 8;
        }
        here = lcode[hold & lmask];
      dolen:
        op = (unsigned)(here.bits);
        hold >>= op;
        bits -= op;
        op = (unsigned)(here.op);
        if (op == 0) {                          /* literal */
            Tracevv((stderr, here.val >= 0x20 && here.val < 0x7f ?
                    "inflate:         literal '%c'\n" :
                    "inflate:         literal 0x%02x\n", here.val));
            *out = (unsigned char)(here.val);
        }
        else if (op & 16) {                     /* length base */
            len = (unsigned)(here.val);
            op &= 15;                           /* number of extra bits */
            if (op) {
                if (__builtin_expect ((bits < op), 0)) {
                    hold = (unsigned long)(*in) << bits;
                    bits = 8;
                }
                len = (unsigned)hold & ((1U << op) - 1);
                hold >>= op;
                bits -= op;
            }
            Tracevv((stderr, "inflate:         length %u\n", len));
            if (bits < 15) {
                hold = (unsigned long)(*in) << bits;
                bits = 8;
                hold = (unsigned long)(*in) << bits;
                bits = 8;
            }
            here = dcode[hold & dmask];
          dodist:
            op = (unsigned)(here.bits);
            hold >>= op;
            bits -= op;
            op = (unsigned)(here.op);
            if (op & 16) {                      /* distance base */
                dist = (unsigned)(here.val);
                op &= 15;                       /* number of extra bits */
                if (bits < op) {
                    hold = (unsigned long)(*in) << bits;
                    bits = 8;
                    if (__builtin_expect ((bits < op), 0)) {
                        hold = (unsigned long)(*in) << bits;
                        bits = 8;
                    }
                }
                dist = (unsigned)hold & ((1U << op) - 1);
#ifdef INFLATE_STRICT
                if (dist > dmax) {
                    strm->msg = (char *)"invalid distance too far back";
                    state->mode = BAD;
                    break;
                }
#endif
                hold >>= op;
                bits -= op;
                Tracevv((stderr, "inflate:         distance %u\n", dist));
                op = (unsigned)(out - beg);     /* max distance in output */
                if (dist > op) {                /* see if copy from window */
                    op = dist - op;             /* distance back in window */
                    if (__builtin_expect ((op > whave), 0)) {
                        if (state->sane) {
                            strm->msg =
                                (char *)"invalid distance too far back";
                            state->mode = BAD;
                            break;
                        }
#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
                        if (len <= op - whave) {
                            do {
                                *out = 0;
                            } while (--len);
                            continue;
                        }
                        len -= op - whave;
                        do {
                            *out = 0;
                        } while (--op > whave);
                        if (op == 0) {
                            from = out - dist;
                            do {
                                *out = *from;
                            } while (--len);
                            continue;
                        }
#endif
                    }
                    from = window;
                    if (__builtin_expect ((wnext == 0), 0)) {           /* very common case */
                        from = wsize - op;
                        if (op < len) {         /* some from window */
                            len -= op;
                            do {
                                *out = *from;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
                    else if (wnext < op) {      /* wrap around window */
                        from = wsize  wnext - op;
                        op -= wnext;
                        if (__builtin_expect ((op < len), 0)) {         /* some from end of window */
                            len -= op;
#pragma loop count (3)
                            do {
                                *out = *from;
                            } while (--op);
                            from = window;
                            if (__builtin_expect ((wnext < len), 0)) {  /* some from start of window */
                                op = wnext;
                                len -= op;
                                do {
                                    *out = *from;
                                } while (--op);
                                from = out - dist;      /* rest from output */
                            }
                        }
                    }
                    else {                      /* contiguous in window */
                        from = wnext - op;
                        if (__builtin_expect ((op < len), 0)) {         /* some from window */
                            len -= op;
#pragma loop count (3)
                            do {
                                *out = *from;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
#pragma loop count (2)
                    while (len > 2) {
                        *out = *from;
                        *out = *from;
                        *out = *from;
                        len -= 3;
                    }
                    if (len) {
                        *out = *from;
                        if (len > 1)
                            *out = *from;
                    }
                }
                else {
#ifndef DONT_USE_EML
                    if (dist <= 4) { /*  заполняем пачками по dist байт сразу и выставляем out */
                        if (len <= 16) {
                            EML_SPEC_MEMSET (out, ((int) dist), ((int) len));
                        }
                        else
                            eml_spec_memset (out, ((int) dist), (int) len);
                        out += len;
                    }
                    else {
#endif /* DONT_USE_EML */
                    from = out - dist;          /* copy direct from output */
//FIXME: if (len > dist) -> out и from пересекаются. Пока так работает.
#pragma unroll (1)
#pragma loop count (2) /* чтобы все было примитивно и никаких оптимизаций */
                        do {                        /* minimum length is three */
                            *out = *from;
                            *out = *from;
                            *out = *from;
                            len -= 3;
                        } while (len > 2);
                        if (len) {
                            *out = *from;
                            if (len > 1)
                                *out = *from;
                        }
#ifndef DONT_USE_EML
                    }
#endif /* DONT_USE_EML */
                }
            }
            else if (__builtin_expect (((op & 64) == 0), 1)) {          /* 2nd level distance code */
                here = dcode[here.val  (hold & ((1U << op) - 1))];
                goto dodist;
            }
            else {
                strm->msg = (char *)"invalid distance code";
                state->mode = BAD;
                break;
            }
        }
        else if (__builtin_expect (((op & 64) == 0), 1)) {              /* 2nd level length code */
            here = lcode[here.val  (hold & ((1U << op) - 1))];
            goto dolen;
        }
        else if (__builtin_expect ((op & 32), 1)) {                     /* end-of-block */
            Tracevv((stderr, "inflate:         end of block\n"));
            state->mode = TYPE;
            break;
        }
        else {
            strm->msg = (char *)"invalid literal/length code";
            state->mode = BAD;
            break;
        }
    } while (in < last && out < end);

    /* return unused bytes (on entry, bits < 8, so in won't go too far back) */
    len = bits >> 3;
    in -= len;
    bits -= len << 3;
    hold &= (1U << bits) - 1;

    /* update state and return */
    strm->next_in = in;
    strm->next_out = out;
    strm->avail_in = (unsigned)(in < last ? 5  (last - in) : 5 - (in - last));
    strm->avail_out = (unsigned)(out < end ?
                                 257  (end - out) : 257 - (out - end));
    state->hold = hold;
    state->bits = bits;
    return;
}

/*
   inflate_fast() speedups that turned out slower (on a PowerPC G3 750CXe):
   - Using bit fields for code structure
   - Different op definition to avoid & for extra bits (do & for table bits)
   - Three separate decoding do-loops for direct, window, and wnext == 0
   - Special case for distance > 1 copies to do overlapped load and store copy
   - Explicit branch predictions (based on measured branch probabilities)
   - Deferring match copy and interspersed it with decoding subsequent codes
   - Swapping literal/length else
   - Swapping window/direct else
   - Larger unrolled copy loops (three is about right)
   - Moving len -= 3 statement into middle of loop
 */

#endif /* !ASMINF */

#endif
#endif

/* inffast.c -- fast decoding
 * Copyright (C) 1995-2017 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

#ifdef DONT_USE_EML

#include "zutil.h"
#include "inftrees.h"
#include "inflate.h"
#include "inffast.h"

#ifdef ASMINF
#  pragma message("Assembler code may have bugs -- use at your own risk")
#else

/*
   Decode literal, length, and distance codes and write out the resulting
   literal and match bytes until either not enough input or output is
   available, an end-of-block is encountered, or a data error is encountered.
   When large enough input and output buffers are supplied to inflate(), for
   example, a 16K input buffer and a 64K output buffer, more than 95% of the
   inflate execution time is spent in this routine.

   Entry assumptions:

        state->mode == LEN
        strm->avail_in >= 6
        strm->avail_out >= 258
        start >= strm->avail_out
        state->bits < 8

   On return, state->mode is one of:

        LEN -- ran out of enough output space or enough available input
        TYPE -- reached end of block code, inflate() to interpret next block
        BAD -- error in block data

   Notes:

    - The maximum input bits used by a length/distance pair is 15 bits for the
      length code, 5 bits for the length extra, 15 bits for the distance code,
      and 13 bits for the distance extra.  This totals 48 bits, or six bytes.
      Therefore if strm->avail_in >= 6, then there is enough input to avoid
      checking for available input while decoding.

    - The maximum bytes that a single length/distance pair can output is 258
      bytes, which is the maximum length that can be coded.  inflate_fast()
      requires strm->avail_out >= 258 for each loop to avoid checking for
      output space.
 */
void ZLIB_INTERNAL inflate_fast(strm, start)
z_streamp strm;
unsigned start;         /* inflate()'s starting value for strm->avail_out */
{
    struct inflate_state FAR *state;
    z_const unsigned char FAR *in;      /* local strm->next_in */
    z_const unsigned char FAR *last;    /* have enough input while in < last */
    unsigned char FAR *out;     /* local strm->next_out */
    unsigned char FAR *beg;     /* inflate()'s initial strm->next_out */
    unsigned char FAR *end;     /* while out < end, enough space available */
#ifdef INFLATE_STRICT
    unsigned dmax;              /* maximum distance from zlib header */
#endif
    unsigned wsize;             /* window size or zero if not using window */
    unsigned whave;             /* valid bytes in the window */
    unsigned wnext;             /* window write index */
    unsigned char FAR *window;  /* allocated sliding window, if wsize != 0 */
    unsigned long hold;         /* local strm->hold */
    unsigned bits;              /* local strm->bits */
    code const FAR *lcode;      /* local strm->lencode */
    code const FAR *dcode;      /* local strm->distcode */
    unsigned lmask;             /* mask for first level of length codes */
    unsigned dmask;             /* mask for first level of distance codes */
    code const *here;           /* retrieved table entry */
    unsigned op;                /* code bits, operation, extra bits, or */
                                /*  window position, window bytes to copy */
    unsigned len;               /* match length, unused bytes */
    unsigned dist;              /* match distance */
    unsigned char FAR *from;    /* where to copy match from */

    /* copy state to local variables */
    state = (struct inflate_state FAR *)strm->state;
    in = strm->next_in;
    last = in + (strm->avail_in - 5);
    out = strm->next_out;
    beg = out - (start - strm->avail_out);
    end = out + (strm->avail_out - 257);
#ifdef INFLATE_STRICT
    dmax = state->dmax;
#endif
    wsize = state->wsize;
    whave = state->whave;
    wnext = state->wnext;
    window = state->window;
    hold = state->hold;
    bits = state->bits;
    lcode = state->lencode;
    dcode = state->distcode;
    lmask = (1U << state->lenbits) - 1;
    dmask = (1U << state->distbits) - 1;

    /* decode literals and length/distances until end-of-block or not enough
       input data or output space */
    do {
        if (bits < 15) {
            hold += (unsigned long)(*in++) << bits;
            bits += 8;
            hold += (unsigned long)(*in++) << bits;
            bits += 8;
        }
        here = lcode + (hold & lmask);
      dolen:
        op = (unsigned)(here->bits);
        hold >>= op;
        bits -= op;
        op = (unsigned)(here->op);
        if (op == 0) {                          /* literal */
            Tracevv((stderr, here->val >= 0x20 && here->val < 0x7f ?
                    "inflate:         literal '%c'\n" :
                    "inflate:         literal 0x%02x\n", here->val));
            *out++ = (unsigned char)(here->val);
        }
        else if (op & 16) {                     /* length base */
            len = (unsigned)(here->val);
            op &= 15;                           /* number of extra bits */
            if (op) {
                if (bits < op) {
                    hold += (unsigned long)(*in++) << bits;
                    bits += 8;
                }
                len += (unsigned)hold & ((1U << op) - 1);
                hold >>= op;
                bits -= op;
            }
            Tracevv((stderr, "inflate:         length %u\n", len));
            if (bits < 15) {
                hold += (unsigned long)(*in++) << bits;
                bits += 8;
                hold += (unsigned long)(*in++) << bits;
                bits += 8;
            }
            here = dcode + (hold & dmask);
          dodist:
            op = (unsigned)(here->bits);
            hold >>= op;
            bits -= op;
            op = (unsigned)(here->op);
            if (op & 16) {                      /* distance base */
                dist = (unsigned)(here->val);
                op &= 15;                       /* number of extra bits */
                if (bits < op) {
                    hold += (unsigned long)(*in++) << bits;
                    bits += 8;
                    if (bits < op) {
                        hold += (unsigned long)(*in++) << bits;
                        bits += 8;
                    }
                }
                dist += (unsigned)hold & ((1U << op) - 1);
#ifdef INFLATE_STRICT
                if (dist > dmax) {
                    strm->msg = (char *)"invalid distance too far back";
                    state->mode = BAD;
                    break;
                }
#endif
                hold >>= op;
                bits -= op;
                Tracevv((stderr, "inflate:         distance %u\n", dist));
                op = (unsigned)(out - beg);     /* max distance in output */
                if (dist > op) {                /* see if copy from window */
                    op = dist - op;             /* distance back in window */
                    if (op > whave) {
                        if (state->sane) {
                            strm->msg =
                                (char *)"invalid distance too far back";
                            state->mode = BAD;
                            break;
                        }
#ifdef INFLATE_ALLOW_INVALID_DISTANCE_TOOFAR_ARRR
                        if (len <= op - whave) {
                            do {
                                *out++ = 0;
                            } while (--len);
                            continue;
                        }
                        len -= op - whave;
                        do {
                            *out++ = 0;
                        } while (--op > whave);
                        if (op == 0) {
                            from = out - dist;
                            do {
                                *out++ = *from++;
                            } while (--len);
                            continue;
                        }
#endif
                    }
                    from = window;
                    if (wnext == 0) {           /* very common case */
                        from += wsize - op;
                        if (op < len) {         /* some from window */
                            len -= op;
                            do {
                                *out++ = *from++;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
                    else if (wnext < op) {      /* wrap around window */
                        from += wsize + wnext - op;
                        op -= wnext;
                        if (op < len) {         /* some from end of window */
                            len -= op;
                            do {
                                *out++ = *from++;
                            } while (--op);
                            from = window;
                            if (wnext < len) {  /* some from start of window */
                                op = wnext;
                                len -= op;
                                do {
                                    *out++ = *from++;
                                } while (--op);
                                from = out - dist;      /* rest from output */
                            }
                        }
                    }
                    else {                      /* contiguous in window */
                        from += wnext - op;
                        if (op < len) {         /* some from window */
                            len -= op;
                            do {
                                *out++ = *from++;
                            } while (--op);
                            from = out - dist;  /* rest from output */
                        }
                    }
                    while (len > 2) {
                        *out++ = *from++;
                        *out++ = *from++;
                        *out++ = *from++;
                        len -= 3;
                    }
                    if (len) {
                        *out++ = *from++;
                        if (len > 1)
                            *out++ = *from++;
                    }
                }
                else {
                    from = out - dist;          /* copy direct from output */
                    do {                        /* minimum length is three */
                        *out++ = *from++;
                        *out++ = *from++;
                        *out++ = *from++;
                        len -= 3;
                    } while (len > 2);
                    if (len) {
                        *out++ = *from++;
                        if (len > 1)
                            *out++ = *from++;
                    }
                }
            }
            else if ((op & 64) == 0) {          /* 2nd level distance code */
                here = dcode + here->val + (hold & ((1U << op) - 1));
                goto dodist;
            }
            else {
                strm->msg = (char *)"invalid distance code";
                state->mode = BAD;
                break;
            }
        }
        else if ((op & 64) == 0) {              /* 2nd level length code */
            here = lcode + here->val + (hold & ((1U << op) - 1));
            goto dolen;
        }
        else if (op & 32) {                     /* end-of-block */
            Tracevv((stderr, "inflate:         end of block\n"));
            state->mode = TYPE;
            break;
        }
        else {
            strm->msg = (char *)"invalid literal/length code";
            state->mode = BAD;
            break;
        }
    } while (in < last && out < end);

    /* return unused bytes (on entry, bits < 8, so in won't go too far back) */
    len = bits >> 3;
    in -= len;
    bits -= len << 3;
    hold &= (1U << bits) - 1;

    /* update state and return */
    strm->next_in = in;
    strm->next_out = out;
    strm->avail_in = (unsigned)(in < last ? 5 + (last - in) : 5 - (in - last));
    strm->avail_out = (unsigned)(out < end ?
                                 257 + (end - out) : 257 - (out - end));
    state->hold = hold;
    state->bits = bits;
    return;
}

/*
   inflate_fast() speedups that turned out slower (on a PowerPC G3 750CXe):
   - Using bit fields for code structure
   - Different op definition to avoid & for extra bits (do & for table bits)
   - Three separate decoding do-loops for direct, window, and wnext == 0
   - Special case for distance > 1 copies to do overlapped load and store copy
   - Explicit branch predictions (based on measured branch probabilities)
   - Deferring match copy and interspersed it with decoding subsequent codes
   - Swapping literal/length else
   - Swapping window/direct else
   - Larger unrolled copy loops (three is about right)
   - Moving len -= 3 statement into middle of loop
 */

#endif /* !ASMINF */
#endif
